//version 1.0 by Marek Mleziva 4/2015
//I2C DMA emulator
//8-bit PIC master write one byte or an sequence of bytes from its RAM into I2C slave chip memory
//slave chip is 7 bit addresed
//or it read one or sequence of bytes from slave chip memory to his RAM
// it use interrupts of PIC I2C module so that program waits no delay cycles
//use control special registers SSPCON1, SSPCON2, SSPBUF, SSPADD and registers for interrupt
//but not use SSPSTAT nor SSPCON3 registers


#include<xc.h>
#define SSP1
//#define SSP2
//#define transfered

#ifdef SSP1
 #define SSPBUF SSP1BUF
 #define SSPADD SSP1ADD
 #define SEN SSP1CON2bits.SEN
 #define RSEN SSP1CON2bits.RSEN
 #define RCEN SSP1CON2bits.RCEN
 #define PEN SSP1CON2bits.PEN
 #define ACKDT SSP1CON2bits.ACKDT
 #define ACKEN SSP1CON2bits.ACKEN
 #define ACKSTAT SSP1CON2bits.ACKSTAT
 #define SSPM SSP1CON1bits.SSPM
 #define SSPEN SSP1CON1bits.SSPEN
 #define SSPIE SSP1IE
 #define SSPIF SSP1IF
 #define BCLIE BCL1IE
 #define BCLIF BCL1IF
#endif

#ifdef SSP2
 #define SSPBUF SSP2BUF
 #define SSPADD SSP2ADD
 #define SEN SSP2CON2bits.SEN
 #define RSEN SSP2CON2bits.RSEN
 #define RCEN SSP2CON2bits.RCEN
 #define PEN SSP2CON2bits.PEN
 #define ACKDT SSP2CON2bits.ACKDT
 #define ACKEN SSP2CON2bits.ACKEN
 #define ACKSTAT SSP2CON2bits.ACKSTAT
 #define SSPM SSP2CON1bits.SSPM
 #define SSPEN SSP2CON1bits.SSPEN
 #define SSPIE SSP2IE
 #define SSPIF SSP2IF
 #define BCLIE BCL2IE
 #define BCLIF BCL2IF
#endif


static signed char step;
static volatile unsigned char *pointerDMAnow,*pointerDMAmax;
volatile struct     //user wouldn't changed it in time of transfer
{
     unsigned Read: 1;  //user set this flag that means I2C is busied by reading
                        //and it is cleared  by routine when I2C transfer finished
     unsigned Write: 1; //the same for writing
     unsigned OK: 1;    //set by routine when I2C transfer finished OK
                        //and cleared when transfer has begining
     unsigned NoWordAdr:1;//when no word address, but only device address
                          //is written(on random access)
     unsigned LongWordAdr:1;//when 2 bytes of word address are used, first byte(wordAdr)
                          //  and second(wordAdr2) - e.g. in serial EEPROM
     void *pointerDMA;    //pointer into RAM of PIC controller, where  data
                          //for writing or by reading  are saved
     unsigned char bytes;// number of data bytes transfered (no address inclusive )
     unsigned char devAdr;  //device address of I2C slave chip
     unsigned char wordAdr; //1 Byte word(register) address in I2C chip
     unsigned char wordAdr2; //2. byte of long word address, if it has 2 Bytes
}IIC;

#ifdef transfered
extern inline void IICtransfered(void);//can be called at the end of IICinterrupt() service,
                                        //when IIC transfer of packet stop.
#endif

//basic part of I2C configuration
inline void IICconfig(char speeddiv)//speeddiv divides maximal I2G clock freq.
{
 step=0;
 SSPM=0b1000;           //I2C master mode
 SSPEN=1; 		//enable I2C serial port
 SSPADD=speeddiv;
 SSPIF=0;
 SSPIE=1;
 BCLIE=1;
}

inline void IICinterrupt(void)
{
 if(BCLIF)
 {
  BCLIF=0;
  IIC.Read=0;
  IIC.Write=0;
  step=0;
  IIC.OK=0;
#ifdef transfered
  IICtransfered();
#endif
 }
 else if(SSPIF)		// if I2C transfer interrupt
 {
  SSPIF=0;
  if(IIC.Read && !IIC.Write)
  {
   switch (step)											//step is I2C communication phase
   {
    case 0: IIC.OK=0;           //I2C start pulse
            pointerDMAnow=(char *)IIC.pointerDMA;
            pointerDMAmax=(char *)IIC.pointerDMA+IIC.bytes;
            if(IIC.NoWordAdr){step=5; SSPBUF=(IIC.devAdr<<1)+1;}
            else {step++;  SSPBUF=IIC.devAdr<<1;}
            break; 
                                                       //slave I2Cchip address for writing
    case 1: if(ACKSTAT){step=-step; PEN=1;}	//stop if no slave receiver
            else{SSPBUF=IIC.wordAdr; step++;
                 if(!IIC.LongWordAdr)step++;
                } 		//Write slave I2Cchip register address
            break;
    case 2: if(ACKSTAT){step=-step; PEN=1;}	//stop if no slave receiver
            else{SSPBUF=IIC.wordAdr2; step++;
                } 		//Write slave I2Cchip register address
            break;
    case 3: if(ACKSTAT){step=-step; PEN=1;}	//stop if no slave receiver
            else {RSEN=1;step++;} break;		// I2C restart pulse if slave detect
    case 4: SSPBUF= (IIC.devAdr<<1)+1;// slave I2Cchip address for reading
            step++; break;
    case 5: RCEN=1; step++; break;		//set IIC.pointerDMA address and receiving
    case 6: *pointerDMAnow=SSPBUF; //Read slave I2Cchip register to memory
            pointerDMAnow++;
            if(pointerDMAnow>=pointerDMAmax){ACKDT=1;step++;} else {ACKDT=0; step--;};
	    ACKEN=1;  break;
    case 7: PEN=1;step++; break;		//I2C stop pulse
    case 8: step=0; IIC.Read=0; IIC.OK=1; break;//I2C reading finished
    default:step=0; IIC.Read=0; IIC.OK=0; break;
   }
  }
  else if(IIC.Write && !IIC.Read)
  {
   switch (step)											//step is I2C communication phase
   {
     case 0: IIC.OK=0;
             pointerDMAnow=(char *)IIC.pointerDMA;
             pointerDMAmax=(char *)IIC.pointerDMA+IIC.bytes;
             if(IIC.NoWordAdr)step=3;
             else step++;
             SSPBUF=IIC.devAdr<<1;
             break;	//I2C start pulse
     case 1: if(ACKSTAT){step=-step;PEN=1;}        	//stop if no slave receiver
             else {SSPBUF=IIC.wordAdr; step++;
                   if(!IIC.LongWordAdr)step++;
                  } 	//Write slave I2Cchip register address
             break;
     case 2: if(ACKSTAT){step=-step;PEN=1;}        	//stop if no slave receiver
             else {SSPBUF=IIC.wordAdr2; step++;
                  } 	//Write slave I2Cchip register address
             break;
     case 3: if(ACKSTAT){step=-step;PEN=1;}        	//stop if no slave receiver
             else if(pointerDMAnow>=pointerDMAmax)
             {
              step++;
              PEN=1;
             }
             else
             {SSPBUF= *pointerDMAnow;  	//Write slave I2Cchip register
              pointerDMAnow++;
             }
             break;
     case 4: step=0; IIC.Write=0; IIC.OK=1;
              break;//I2C reading finished
    default: step=0; IIC.Write=0; IIC.OK=0; break;
   }
  }
  else
  {
   if(step < 0){step=0; IIC.Write=0; IIC.Read=0; IIC.OK=0;}
   else {step= -1; PEN=1;}
  }
  #ifdef transfered
  if(!IIC.Read && !IIC.Write)
   IICtransfered();
  #endif
 }

}